package Opt;

##
##	This is the `Opt.pm' module for reading parameters from command 
##  line and file file.
##
##	Copyright (c) 2001 Jiri Polcar. All rights reserved.
##	<polcar@physics.muni.cz>
##
##	This program is free software; you can redistribute it and/or
##	modify it under GNU GPL.
##
##

use Exporter;
use Getopt::Long;

use strict;

@main::ISA = qw(Exporter);
@main::EXPORT_OK = qw( getOptionsFromCmdLine  
                       getOptionsFromFile 
                       simplifyOptions
                     );

my $true = 1;
my $false = undef;
my $parsing_err = 0;
my $not_a_file = undef;


# read options from command line; 
sub getOptionsFromCmdLine {
	my ( $opt ) = @_;

#
# in `$opt' is already configurated (default, or so)
#

	my @to_get_options;
	my $key;
	my $realy_param;
	my %options;


	@to_get_options = ();

#
# Prepare input parametrs for using `GetOptions()'
#

	foreach $key ( keys( %{$opt} ) ) {
	  push( @to_get_options, $key.$opt->{$key}->{alias}.$opt->{$key}->{param});
	}
	Getopt::Long::Configure ("bundling");
	unless ( GetOptions( \%options, @to_get_options ) ) { return $parsing_err; }
  
  #care about NEGATION
  # it is BOOLEAN option, return TRUE; must care about possible negation

  my $neg;
  foreach  $key (  keys %options ) {
    if ( exists( $opt->{$key}->{negateto} )) { 
      $neg = $opt->{$key}->{negateto};
      unless ( exists( $opt->{$neg} ) ) {
        	print STDERR "ERROR: Can not negate `".$key."' option!\n";
          return $parsing_err;
      } else {
        # must update directly `$opt', because in `updateDefaultsOptions()'
        # are testes `defined()'
        $options{$neg} = undef;
        $options{$key} = $true;
      }
    }
  }

#
# we have TWO options hash array %{$opt} and %options
# must me "transparent"
#
#	%options	:	siple hash with new values from
#             Getopt::Long::GetOptions
#	%{$opt}  	:	large definition...
#
	updateDefaultsOptions( \%options, $opt );	
	return $true;
}


#	read options ( `$opt' ) from config file `$file'; can read scalar, array and
#	hash value; works only with single lines!!!
sub getOptionsFromFile {
	my ( $file, $opt ) = @_;


	unless ( open( C_FILE, "<$file" ) ) {
		return $not_a_file;
	}
  my $to_parse;
  my $key_word;
  my $o_value;
  my $type;
	my $line;

	my %options;

	while ( $line = <C_FILE> ) {
    chomp( $line ); 
    # discard comments and empty lines
    if ( ($line =~ /^\s*([^\#]*)/) && ( $1 ne '') ) {
			$to_parse = $1;
      # try to get key word and value
      if ( $to_parse =~ /^\s*([^\s=]+)\s*=?\s*(.*)?$/ ) {
        $key_word = $1;
        $o_value = $2;
        # is it realy defined option kwy-word?
			  if ( exists( $opt->{$key_word} ) ) {
          # switch due to TYPE of option: ARRAY, HASH, SCALAR
          $type = ref( $opt->{$key_word}->{value} );

careAboutOption( $opt, $file, $key_word,$o_value,$type,$to_parse,\%options );

			  } else { 
          print STDERR "$file: WARNING: Error  at $. line.\n"; 
          return $parsing_err;
        }
		  }  else { 
        print STDERR "$file: Can not get a KEY WORD on line $..\n"; 
        return $parsing_err;
      } 
	  } # else  { print STDERR "$file: Line $. is a comment or empty.\n"; }
  }
	close( C_FILE );
	updateDefaultsOptions( \%options, $opt );	
	return $true;
}

sub careAboutOption  {
  my ( $opt, $file, $key_word, $o_value, $type, $to_parse, $options) = @_;

  my $value;
  my $key;

# ARRAY

  if ( $type =~ /ARRAY/ )	{
	  unless ( defined( $value = 
             uScalar( $o_value, $key_word, $file, $opt, $options ) ) ) {
      return $parsing_err;
		} else { push( @{$options->{$key_word}}, $value ); }	

# HASH

  } elsif  ( $type =~ /HASH/ )	{
	  ( $key, $value ) = uHash( $o_value, $file, $opt );
		unless ( defined( $key ) && defined( $value ) ) {
      return $parsing_err;
    } else {  ${$options->{$key_word}}{$key} =  $value; }

# SCALAR

  }  else	{ 
    unless ( defined( $options->{$key_word} = 
             uScalar( $o_value, $key_word,  $file, $opt, $options ) ) ) {
      return $parsing_err;
    }
  }

}

#	needy to `getOptionsFromFile'; read scalar value; return UNDEF if
#	unsuccessfull
sub uScalar {
	my ( $o_value, $key_word, $file, $opt, $options ) = @_;

  my $value;
  my $neg;

  # is is BOOLEAN option? 
	if ( $opt->{$key_word}->{param} =~ /\=/ ) {

    # is it a integer option?
	  if ( $opt->{$key_word}->{param} =~ /i/i ) {
			unless ( $o_value =~ /^[-+]?\d+$/) {
				print STDERR "$file: ERROR: Expect integer value at line $..\n";
				return $parsing_err;
			} else { return $o_value; }

    # is it a float option?
		} elsif ( $opt->{$key_word}->{param} =~ /f/i ) {
			unless ( $o_value =~ /^[-+]?\d*\.?\d*([eE][-+]?\d)?\d*$/) {
				print STDERR "$file: ERROR: Expect float value at line $..\n";
				return $parsing_err;
			} else { return $o_value; }

    # it will be a string option
		} else { return $o_value;}

	} else { 
    # it is BOOLEAN option, return TRUE; must care about possible negation
    if ( exists( $opt->{$key_word}->{negateto} )) { 
      $neg = $opt->{$key_word}->{negateto};
      unless ( exists( $opt->{$neg} ) ) {
        	print STDERR "$file: ERROR: Can not negate `".$neg."' option!\n";
          return $parsing_err;
      } else {        
        $options->{$neg} = undef;
      }
    }
    return $true; 
  }

}

#	needy to `getOptionsFromFile'; read hash value; return (UNDEF, UNDEF) if
#	unsuccess
sub uHash {
	my ( $o_value, $file, $opt ) = @_;

	my $key = undef; 
	my $value = undef;

  if ( $o_value =~ /^\s*([^\s=]+)\s*=?\s*(.*)?$/ ) {
	  $key =$1;
	  $value = $2;
	}
	unless ( defined( $key) && defined( $value ) ) {
		print STDERR "$file: ERROR: ".
                  "Hash parametr expects KEY and VALUE at line $..\n";
		return (undef, undef);
	} else {
		return ( $key, $value );
	}
}


sub updateDefaultsOptions {
	my ( $options, $opt ) = @_;			#	options	:	siple hash with new values from
                                  #           Getopt::Long::GetOptions
																	#	opt			:	large definition...
	my $key;
	my $tmp;
	my $something;

	foreach $key (keys %{$opt} ) {
# BOOLEAN
		if ( exists( $options->{$key} ) ) {
			$something = $false;
			$_ = ref($options->{$key});

			if ( /HASH/ ) {	
				foreach $tmp ( keys %{$options->{$key}} ) {
          $opt->{$key}->{value}->{$tmp} = $options->{$key}->{$tmp};
        }
			}	elsif ( /ARRAY/ ) {	
        # default value for DUPLICITE
        unless ( defined ( $opt->{$key}->{duplicite} )) { 
          $opt->{$key}->{duplicite} = $false; 
        }
				foreach $tmp ( @{$options->{$key}} ) {
          unless ( $opt->{$key}->{duplicite} ) {
            unless ( isIn( $opt->{$key}->{param} , $tmp , 
                           @{$opt->{$key}->{value}} ) ) {
              push( @{$opt->{$key}->{value}}, $tmp );
            }
          } else {
            push( @{$opt->{$key}->{value}}, $tmp );
          }
        }
			} else {
				$opt->{$key}->{value} = $options->{$key}; 
			}
		} else { next; } 
	}
	return $true;
}


sub isIn {
  my ( $param, $element, @array ) = @_;
  my $tmp;
  foreach $tmp ( @array ) {
    if ( $param =~ /=[if]/ ) {
      if ( $element == $tmp ) { return $true;}
    } else {
      if ( $element eq $tmp ) { return $true;}
    }
  }
  return $false;
}



sub simplifyOptions {
	my ( $opt ) = @_;

	my $options = {};
	my $key;

	foreach $key ( keys %{$opt}) { $options->{$key} = $opt->{$key}->{value}; }
	return $options;
}



$true;
